import json
import time
import copy
from dataclasses import dataclass
from typing import Optional, Union, Any

import requests
from requests.exceptions import ConnectionError    # noqa A004

from metasdk.logger import LOGGER_ENTITY
from metasdk.exceptions import RetryHttpRequestError, EndOfTriesError, UnexpectedError, ApiProxyError, RateLimitError, \
    ApiProxyBusinessErrorMixin


@dataclass
class RequestPolicy:
    """Политика выполнения HTTP-запросов: таймауты и повторные попытки."""

    max_retries: int = 20
    """Максимальное количество попыток."""

    retry_delay_sec: int = 20
    """Задержка между попытками в секундах."""

    timeout_sec: int | float = 3600
    """Общий таймаут запроса в секундах (используется если connect/read не заданы)."""

    connect_timeout_sec: Optional[int | float] = None
    """Таймаут установки соединения. None означает без лимита (если задан read_timeout_sec)."""

    read_timeout_sec: Optional[int | float] = None
    """Таймаут чтения ответа. None означает без лимита (если задан connect_timeout_sec)."""

    @property
    def timeout(self) -> Union[int | float, tuple[int | float | None, int | float | None]]:
        """
        Возвращает настройки таймаута для requests.

        - Если оба connect/read не заданы — возвращает timeout_sec (общий таймаут).
        - Если хотя бы один задан — возвращает tuple (connect, read).
          None в tuple интерпретируется requests как "без таймаута".
        """
        if self.connect_timeout_sec is None and self.read_timeout_sec is None:
            return self.timeout_sec

        return self.connect_timeout_sec, self.read_timeout_sec


class ApiProxyService:
    MAX_PAGES = 100000

    def __init__(self, app):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options = {}
        self.__data_get_cache = {}
        self.__data_get_flatten_cache = {}

    def __api_proxy_call(
        self,
        engine: str,
        payload: dict[str, Any],
        method: str,
        analyze_json_error_param: bool,
        retry_request_substr_variants: list[str],
        stream: bool = False,
        raise_business_errors: bool = False,
        log: bool = False,
        request_policy: RequestPolicy | None = None,
    ):
        """
        :param engine: Система
        :param payload: Данные для запроса
        :param method: string Может содержать native_call | tsv | json_newline
        :param analyze_json_error_param: Нужно ли производить анализ параметра error в ответе прокси
        :param retry_request_substr_variants: Список подстрок, при наличии которых в ответе будет происходить перезапрос
        :param stream: стримиговый запрос, можно будет итерироваться по строкам ответа, не загружая весь ответ в память
        :param log: логировать параметры запроса
        :param request_policy: Политика выполнения запроса (таймауты, ретраи). По умолчанию RequestPolicy().
        :return:
        """
        policy = request_policy or RequestPolicy()

        log_ctx = {
            "engine": engine,
            "method": payload.get('method'),
            "method_params": payload.get('method_params')
        }
        if log:
            self.__app.log.info("Call api proxy", log_ctx)
        body = {
            "engine": engine,
            "payload": payload
        }
        for _try_idx in range(policy.max_retries):
            try:
                body_str = json.dumps(body)
                headers = {
                    "User-Agent": self.__app.user_agent,
                    "X-App": "META",
                    "X-Worker": self.__app.service_id,
                    "X-ObjectLocator": LOGGER_ENTITY.get("objectLocator")
                }
                resp = requests.post(
                    self.__app.api_proxy_url + "/" + method,
                    body_str,
                    timeout=policy.timeout,
                    stream=stream,
                    headers=headers,
                )

                self.check_err(resp, analyze_json_error_param=analyze_json_error_param,
                               retry_request_substr_variants=retry_request_substr_variants,
                               raise_business_errors=raise_business_errors)
                return resp
            except (RetryHttpRequestError, RateLimitError, ConnectionError) as e:
                self.__app.log.warning("Sleep retry query: " + str(e), log_ctx)
                sleep_time = policy.retry_delay_sec

                if e.__class__.__name__ == "RateLimitError":
                    sleep_time = e.waiting_time

                time.sleep(sleep_time)
        raise EndOfTriesError("Api of api proxy tries request")

    def call_proxy_with_paging(
        self,
        engine: str,
        payload: dict[str, Any],
        method: str,
        analyze_json_error_param: bool,
        retry_request_substr_variants: list[str],
        max_pages: int = MAX_PAGES,
        raise_business_errors: bool = False,
        log: bool = False,
        request_policy: RequestPolicy | None = None,
    ):
        """
        Постраничный запрос
        :param engine: Система
        :param payload: Данные для запроса
        :param method: string Может содержать native_call | tsv | json_newline
        :param analyze_json_error_param: Нужно ли производить анализ параметра error в ответе прокси
        :param retry_request_substr_variants: Список подстрок, при наличии которых в ответе будет происходить перезапрос
        :param max_pages: Максимальное количество страниц в запросе
        :param raise_business_errors: Преобразовывать ApiProxyError в конкретную BusinessError
        :param log: логировать параметры запроса
        :param request_policy: Политика выполнения запроса (таймауты, ретраи). По умолчанию RequestPolicy().
        :return: объект генератор
        """
        copy_payload = copy.deepcopy(payload)

        idx = 0
        for idx in range(max_pages):    # noqa B007
            resp = self.__api_proxy_call(
                engine,
                copy_payload,
                method,
                analyze_json_error_param,
                retry_request_substr_variants,
                raise_business_errors=raise_business_errors,
                log=log,
                request_policy=request_policy,
            )
            yield resp

            paging_resp = resp.json().get("paging")
            if not paging_resp:
                break
            copy_payload["paging"] = paging_resp

        if idx >= max_pages:
            self.__app.log.warning("Достигнут максимальный предел страниц", {"max_pages": max_pages})

    def call_proxy(
        self,
        engine: str,
        payload: dict[str, Any],
        method: str,
        analyze_json_error_param: bool,
        retry_request_substr_variants: list[str],
        stream: bool = False,
        raise_business_errors: bool = False,
        log: bool = False,
        request_policy: RequestPolicy | None = None,
    ):
        """
        :param engine: Система
        :param payload: Данные для запроса
        :param method: string Может содержать native_call | tsv | json_newline
        :param analyze_json_error_param: Нужно ли производить анализ параметра error в ответе прокси
        :param retry_request_substr_variants: Список подстрок, при наличии которых в ответе будет происходить перезапрос
        :param raise_business_errors: Преобразовывать ApiProxyError в конкретную BusinessError
        :param stream: стримиговый запрос, можно будет итерироваться по строкам ответа, не загружая весь ответ в память
        :param log: логировать параметры запроса
        :param request_policy: Политика выполнения запроса (таймауты, ретраи). По умолчанию RequestPolicy().
        :return:
        """
        return self.__api_proxy_call(
            engine,
            payload,
            method,
            analyze_json_error_param,
            retry_request_substr_variants,
            stream,
            raise_business_errors=raise_business_errors,
            request_policy=request_policy,
        )

    @staticmethod
    def check_err(resp, analyze_json_error_param=False, retry_request_substr_variants=None,
                  raise_business_errors=False):
        """
        :type retry_request_substr_variants: list Список вхождений строк, при налиции которых в ошибке апи будет произведен повторный запрос к апи
        """
        if retry_request_substr_variants is None:
            retry_request_substr_variants = []

        # РКН блокировки вызывают ошибку SSL
        retry_request_substr_variants.append("TLSV1_ALERT_ACCESS_DENIED")

        if resp.status_code in [502, 503, 504]:
            raise RetryHttpRequestError(resp.text)

        if resp.status_code >= 400:
            rtext = resp.text
            for v_ in retry_request_substr_variants:
                if v_ in rtext:
                    raise RetryHttpRequestError(rtext)
            raise UnexpectedError("HTTP request failed: {} {}".format(resp.status_code, rtext))
        if analyze_json_error_param:
            data_ = resp.json()
            error = data_.get("error")
            if error:
                full_err_ = json.dumps(error, ensure_ascii=False)

                if error.get("type") == "RateLimitError":
                    raise RateLimitError(error.get("message"), waiting_time=error.get("waiting_time"))

                for v_ in retry_request_substr_variants:
                    if v_ in full_err_:
                        raise RetryHttpRequestError(full_err_)
                if raise_business_errors:
                    for cls in ApiProxyBusinessErrorMixin.__subclasses__():
                        if cls.__name__ == error.get("type"):
                            raise cls(full_err_)
                raise ApiProxyError(full_err_)
        return resp



import json
import runpy
import os
import argparse

from metasdk import MetaApp
from metasdk.__state import get_current_app


def main():
    """
    Запускает указанный МЕТА скрипт и извлекает из него конструкции @META.event_bus.listener,
    которые используются для объявления функции как обработчика конкретных типов событий
    """
    parser = argparse.ArgumentParser()
    parser.add_argument('--path_name', help='Full absolute filepath', type=str, required=True)
    args = parser.parse_args()

    build_event_list(args.path_name)


def build_event_list(path_name):
    os.environ['GARPUN_CLOUD_GENERATE'] = 'y'
    runpy.run_path(path_name=path_name)

    META: MetaApp = get_current_app()

    res = []
    for listener in META.event_bus.get_listeners():
        res.append(listener.listen.__dict__)

    print('%START_GARPUN_CLOUD_GENERATE%\n' + json.dumps(res) + '\n%END_GARPUN_CLOUD_GENERATE%')    # noqa: T201    


if __name__ == '__main__':
    main()



import time


class BulkLogger:
    def __init__(self, log, log_message, total=None, part_log_time_minutes=1):
        self.__log = log
        self.__begin_time = time.time()

        self.__log_message = log_message
        self.__part_log_time_seconds = part_log_time_minutes * 60
        self.__counter = 0

        if total is not None:
            if total <= 0:
                self.__log.info('Нет элементов для логирования. Вероятно список массив пустой')
                total = None

        self.__total = total
        self.__percent_done = 0

    def try_log_part(self, context=None, with_start_message=True):
        """
        Залогировать, если пришло время из part_log_time_minutes
        :return: boolean Возвращает True если лог был записан
        """
        if context is None:
            context = {}
        self.__counter += 1
        if time.time() - self.__begin_time > self.__part_log_time_seconds:
            self.__begin_time = time.time()
            context['count'] = self.__counter
            if self.__total:
                self.__percent_done = int(self.__counter * 100 / self.__total)
                context['percentDone'] = self.__percent_done
                context['total'] = self.__total
            self.__log.info(msg=self.__log_message, context=context)
            return True
        elif self.__counter == 1:
            if with_start_message:
                self.__log.info(u"Начали цикл: " + self.__log_message)
            return True
        return False

    def finish(self):
        self.__log.info(u"Закончили цикл: " + self.__log_message)

    def get_counter(self):
        return self.__counter

    def get_percent_done(self):
        return self.__percent_done



import asyncio
import json
from typing import Any, BinaryIO, Mapping, Optional, Sequence, TYPE_CHECKING


if TYPE_CHECKING:
    from metasdk import MetaApp
    from metasdk.transport import AsyncNativeApiTransport


class AsyncDbQueryService:
    """Асинхронный клиент для работы с БД."""

    def __init__(
        self,
        app: "MetaApp",
        options: Mapping[str, Any],
        transport: Optional["AsyncNativeApiTransport"] = None,
        owns_transport: bool = False,
    ) -> None:
        """
        Инициализирует асинхронный сервис работы с БД.

        :param app: Экземпляр MetaApp.
        :param options: Опции подключения (dbAlias, shardKey и др.).
        :param transport: Опциональный транспорт. Если не передан — берём у app.
        :param owns_transport: Закрывать ли транспорт при выходе из контекста.
        :return: None.
        """
        self.__app = app
        self.__options = options
        self.__transport = transport
        self.__owns_transport = owns_transport

    async def __aenter__(self) -> "AsyncDbQueryService":
        self._ensure_transport()
        return self

    async def __aexit__(self, exc_type, exc, tb) -> None:
        if self.__owns_transport and self.__transport:
            await self.__transport.aclose()
            self.__transport = None

    def _ensure_transport(self) -> "AsyncNativeApiTransport":
        if self.__transport:
            return self.__transport

        # Берём кешированный транспорт из MetaApp.
        if self.__app:
            self.__transport = self.__app.ensure_async_transport()
            return self.__transport

        raise ValueError("Async transport is not configured")

    async def schema_data(self, configuration: Mapping[str, Any]) -> dict:
        """
        Получает схему данных по конфигурации.

        :param configuration: Конфигурация запроса схемы.
        :return: Словарь с данными схемы.
        """
        params = {"configuration": json.dumps(configuration)}
        transport = self._ensure_transport()
        response = await transport.call("db", "schema-data", params, self.__options, multipart_form=True)
        return await self._json_response(response)

    async def upload_data(self, file_descriptor: BinaryIO, configuration: Mapping[str, Any]) -> dict:

        raise NotImplementedError("upload_data is temporarily disabled for async client")

    async def download_data(self, configuration: Mapping[str, Any], output_file: str) -> None:
        raise NotImplementedError("download_data is temporarily disabled for async client")

    async def batch_update(self, command: str, rows: Sequence[Mapping[str, Any]]) -> dict[str, Any]:
        """
        Массовая вставка/обновление записей (рекомендуется для 1-5к записей за вызов).

        :param command: SQL-команда INSERT или UPDATE.
        :param rows: Список словарей с данными строк.
        :return: Словарь с результатом операции.
        """
        request = {
            "database": {
                "alias": self.__options["dbAlias"]
            },
            "batchUpdate": {
                "command": command,
                "rows": rows,
                "shardKey": self.__options.get("shardKey"),
            }
        }
        transport = self._ensure_transport()
        response = await transport.call("db", "batch-update", request, self.__options, multipart_form=False)
        return await self._json_response(response)

    async def update(self, command: str, params: Optional[Mapping[str, Any]] = None) -> dict[str, Any]:
        """
        Выполняет запросы INSERT, UPDATE, DELETE и пр., не возвращающие результата.
        Исключение — запросы с RETURNING для PostgreSQL.

        :param command: SQL-запрос.
        :param params: Параметры для prepared statements.
        :return: Словарь с результатом операции (DataResult).
        """
        request = {
            "database": {
                "alias": self.__options["dbAlias"]
            },
            "dbQuery": {
                "command": command,
                "parameters": params,
                "shardKey": self.__options.get("shardKey"),
            }
        }
        transport = self._ensure_transport()
        response = await transport.call("db", "update", request, self.__options, multipart_form=False)
        return await self._json_response(response)

    async def query(self, command: str, params: Optional[Mapping[str, Any]] = None, max_rows: int = 0) -> dict[str, Any]:
        """
        Выполняет запрос, который ОБЯЗАТЕЛЬНО должен вернуть результат.
        Для INSERT, UPDATE, DELETE используйте метод update
        или возвращайте результат через конструкцию RETURNING (недоступно в MySQL).

        Пример:
            await db.query('SELECT * FROM users WHERE id=:id', {"id": MY_USER_ID})

        :param command: SQL-запрос.
        :param params: Параметры для prepared statements.
        :param max_rows: Максимальное кол-во строк. Если запрос вернёт больше — ошибка.
                         При 0 действуют стандартные ограничения (50000).
        :return: Словарь с результатом (DataResult).
        """
        request = {
            "database": {
                "alias": self.__options["dbAlias"]
            },
            "dbQuery": {
                "maxRows": max_rows,
                "command": command,
                "parameters": params,
                "shardKey": self.__options.get("shardKey"),
            }
        }
        transport = self._ensure_transport()
        response = await transport.call("db", "query", request, self.__options, multipart_form=False)
        return await self._json_response(response)

    async def one(self, command: str, params: Optional[Mapping[str, Any]] = None) -> Optional[dict[str, Any]]:
        """
        Возвращает первую строку ответа, полученного через query.

        Пример:
            await db.one('SELECT * FROM users WHERE id=:id', {"id": MY_USER_ID})

        :param command: SQL-запрос.
        :param params: Параметры для prepared statements.
        :return: Словарь с данными первой строки или None, если результат пуст.
        """
        data = await self.query(command, params)
        rows = data.get("rows", [])
        return rows[0] if rows else None

    async def all(self, command: str, params: Optional[Mapping[str, Any]] = None) -> list[dict[str, Any]]:  # noqa: A003
        """
        Возвращает все строки ответа, полученного через query.

        Пример:
            await db.all('SELECT * FROM users WHERE id=:id', {"id": MY_USER_ID})

        :param command: SQL-запрос.
        :param params: Параметры для prepared statements.
        :return: Список словарей с данными строк.
        """
        data = await self.query(command, params)
        return data.get("rows", [])

    async def _json_response(self, response) -> dict[str, Any]:
        """
        Читает тело ответа и декодирует JSON.

        :param response: HTTP-ответ (httpx.Response).
        :return: Словарь с декодированными данными.
        """
        try:
            body = await response.aread()
            text = body.decode(response.encoding or "utf-8")
            return json.loads(text)
        finally:
            await response.aclose()

    async def _write_stream_to_file(self, response, output_file: str) -> None:
        """
        Записывает потоковый ответ в файл.

        :param response: HTTP-ответ (httpx.Response) в режиме stream.
        :param output_file: Путь к файлу для записи.
        """
        loop = asyncio.get_running_loop()
        with open(output_file, "wb") as out_file:
            async for chunk in response.aiter_bytes():
                if not chunk:
                    continue
                await loop.run_in_executor(None, out_file.write, chunk)



from __future__ import print_function
import atexit
import logging
import sys
import traceback
from contextvars import ContextVar
from typing import Any, Iterator

from fluent import handler

# http://stackoverflow.com/questions/11029717/how-do-i-disable-log-messages-from-the-requests-library
# Отключаем логи от бибилиотеки requests
logging.getLogger("requests").setLevel(logging.WARNING)
logging.getLogger("urllib3").setLevel(logging.WARNING)


class ContextVarDict:
    """Dict-like обёртка над ContextVar.

    В sync-контексте работает как глобальный dict.
    В async-контексте каждая Task получает изолированный dict.
    """

    def __init__(self, name: str):
        self._var: ContextVar[dict[str, Any] | None] = ContextVar(name, default=None)

    def _get_dict(self) -> dict[str, Any]:
        d = self._var.get()
        if d is None:
            d = {}
            self._var.set(d)
        return d

    def __getitem__(self, key: str) -> Any:
        return self._get_dict()[key]

    def __setitem__(self, key: str, value: Any) -> None:
        self._get_dict()[key] = value

    def __delitem__(self, key: str) -> None:
        del self._get_dict()[key]

    def __contains__(self, key: object) -> bool:
        return key in self._get_dict()

    def __bool__(self) -> bool:
        return bool(self._get_dict())

    def __iter__(self) -> Iterator[str]:
        return iter(self._get_dict())

    def __len__(self) -> int:
        return len(self._get_dict())

    def get(self, key: str, default: Any = None) -> Any:
        return self._get_dict().get(key, default)

    def pop(self, key: str, *default: Any) -> Any:
        return self._get_dict().pop(key, *default)

    def update(self, *args, **kwargs) -> None:
        self._get_dict().update(*args, **kwargs)

    def clear(self) -> None:
        self._get_dict().clear()

    def keys(self):
        return self._get_dict().keys()

    def values(self):
        return self._get_dict().values()

    def items(self):
        return self._get_dict().items()

    def setdefault(self, key: str, default: Any = None) -> Any:
        return self._get_dict().setdefault(key, default)


# Основные контейнеры — автоматически изолируются в async через ContextVar
LOGGER_ENTITY: ContextVarDict = ContextVarDict('logger_entity')
REQUEST_LOG: ContextVarDict = ContextVarDict('request_log')


def init_request_context():
    """Создаёт новый изолированный контекст для текущего request/task.

    Вызывать в начале обработки каждого HTTP request (например, в aiohttp middleware).
    Создаёт новые dict'ы в ContextVar, изолируя данные от других task'ов.
    """
    LOGGER_ENTITY._var.set({})
    REQUEST_LOG._var.set({})


def eprint(*args, **kwargs):
    print(*args, file=sys.stderr, **kwargs)    # noqa: T201


def create_logger(service_id=None, service_ns=None, build_num=None, gcloud_log_host_port=None, debug=True):
    if not service_id:
        service_id = 'unknown'

    # http://stackoverflow.com/questions/3220284/how-to-customize-the-time-format-for-python-logging
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    ch = logging.StreamHandler(sys.stdout)
    formatter = StdoutFormatter('%(asctime)s:%(levelname)s: %(message)s %(context)s', datefmt="%H:%M:%S")
    ch.setFormatter(formatter)

    root_logger.addHandler(ch)

    if not debug:
        gcloud_logs_parts = gcloud_log_host_port.split(':')
        if len(gcloud_logs_parts) != 2:
            raise ValueError("GCLOUD_LOG_HOST_PORT задан неправильно. Проверьте правильность написания HOST:PORT")

        service = service_ns + '.' + service_id
        h = handler.FluentHandler(service, host=gcloud_logs_parts[0], port=int(gcloud_logs_parts[1]))

        g_cloud_formatter = GCloudFormatter()
        g_cloud_formatter.service = service
        g_cloud_formatter.build_num = build_num
        h.setFormatter(g_cloud_formatter)

        root_logger.addHandler(h)

        def exit_handler():
            # Обязательно закрыть сокет по доке
            # https://github.com/fluent/fluent-logger-python
            h.close()

        atexit.register(exit_handler)


def prepare_errors(record):
    dict__ = record.__dict__

    dict__.setdefault('context', {})
    context = dict__.get('context')
    context.update(LOGGER_ENTITY)

    try:
        if record.levelno >= 40:
            if REQUEST_LOG:
                context["httpRequest"] = dict(REQUEST_LOG)
            # Ошибки и выше готовим для Google Cloud ErrorReporting
            # https://cloud.google.com/error-reporting/reference/rest/v1beta1/ErrorContext
            exc_type, exc_value, exc_tb = sys.exc_info()
            a = traceback.extract_tb(exc_tb)
            first_ex = a[-1]
            report_location = {'filePath': first_ex.filename, 'functionName': first_ex.name,
                               'lineNumber': first_ex.lineno}
            context.update({
                'reportLocation': report_location,
            })
    except Exception:
        pass

    ex = context.get("e")
    if ex:
        context.update({
            'e': {
                'message': str(ex),
                'trace': str(traceback.format_exc()),
            }}
        )
    return context


class GCloudFormatter(handler.FluentRecordFormatter, object):
    service = None
    build_num = None

    def format(self, record):    # noqa: A003
        # Create message dict
        context = record.context if hasattr(record, 'context') else {}
        context.update(LOGGER_ENTITY)
        context.update(prepare_errors(record))

        message = {
            "message": record.getMessage(),
            "severity": record.levelname,
            "serviceContext": {
                "service": self.service,
                "version": self.build_num
            }
        }
        if 'e' in context and isinstance(context['e'], dict) and 'trace' in context['e']:
            message['stack_trace'] = message['message'] + "\n" + context['e']['trace']
            # получение имени класса ошибки, автоматически добавляется entity
            message['error_class'] = message['message'].split()[-1]
            del context['e']['trace']
        message['context'] = context

        return message

    def formatException(self, record):
        """
        Format and return the specified exception information as a string.
        :type record logging.LogRecord
        :rtype: dict
        """
        if record.exc_info is None:
            return {}

        (exc_type, exc_message, trace) = record.exc_info

        return {
            'e': {
                'class': str(type(exc_type).__name__),  # ZeroDivisionError
                'message': str(exc_message),  # integer division or modulo by zero
                'trace': list(traceback.format_tb(trace)),
            }
        }


class StdoutFormatter(logging.Formatter, object):
    def format(self, record):    # noqa: A003
        record.message = record.getMessage()
        if self.usesTime():
            record.asctime = self.formatTime(record, self.datefmt)
        dict__ = record.__dict__
        dict__.setdefault('context', {})
        context = dict__.get('context')
        context.update(LOGGER_ENTITY)
        ex = context.get("e")
        if ex:
            context.update({'e': {
                'class': str(type(ex).__name__),
                'message': str(ex),
                'trace': str(traceback.format_exc()),
            }})
            eprint(str(ex) + "\n" + str(traceback.format_exc()))
        s = self._fmt % dict__
        if record.exc_info:
            # Cache the traceback text to avoid converting it multiple times
            # (it's constant anyway)
            if not record.exc_text:
                record.exc_text = self.formatException(record.exc_info)
        if record.exc_text:
            if s[-1:] != "\n":
                s = s + "\n"
            try:
                s = s + record.exc_text
            except UnicodeError:
                # Sometimes filenames have non-ASCII chars, which can lead
                # to errors when s is Unicode and record.exc_text is str
                # See issue 8924.
                # We also use replace for when there are multiple
                # encodings, e.g. UTF-8 for the filesystem and latin-1
                # for a script. See issue 13232.
                s = s + record.exc_text.decode(sys.getfilesystemencoding(),
                                               'replace')
        return s



import json


class UserManagementService:
    def __init__(self, app):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options = {}

    def send_recovery_notice(self, login, app_alias, state=None):
        """
        Выслать письмо о восстановлении пароля
        """
        data = {
            "login": login,
            "appAlias": app_alias,
            "state": state,
        }
        response = self.__app.native_api_call(
            service="user-management",
            method="sendRecoveryNotice",
            data=data,
            options=self.__options,
            multipart_form=False,
            multipart_form_data=None,
            stream=False,
            http_path="/api/meta/v1/",
            http_method="POST",
        )
        return json.loads(response.text)



import json


class ExportService:
    def __init__(self, app):
        """
        :type app: metasdk.MetaApp
        """
        self.__app = app
        self.__options = {}

    def export_page(self, app_id, page_id, entity_id, object_id, export_data_ids, params=None, export_format="csv", export_empty=False):
        if params is None:
            params = {"stateParams": {}}
        response = self.__app.native_api_call('export', 'page', params, self.__options, get_params={
            "applicationId": app_id,
            "entityPageId": page_id,
            "entityId": entity_id,
            "objectId": object_id,
            "exportFormat": export_format,
            "exportEmpty": export_empty,
            "exportDataIds": export_data_ids,
        })
        return json.loads(response.text)

    def export_data_source(self, ds_id):
        params = {}
        response = self.__app.native_api_call('export', 'data_source', params, self.__options, get_params={
            "dataSourceId": ds_id,
        })
        return json.loads(response.text)



import asyncio
import json
from typing import Any, Callable, Optional

import httpx

from metasdk.exceptions import ServerError
from metasdk.logger.logger import Logger
from metasdk.services import process_meta_api_error_code


class AsyncNativeApiTransport:
    """
    Независимый от MetaApp асинхронный транспорт для вызова native API.

    Использует httpx.AsyncClient для выполнения HTTP-запросов с поддержкой
    повторных попыток при сетевых ошибках и ошибках сервера.
    """

    def __init__(
        self,
        meta_url: str,
        default_headers: dict[str, str],
        logger: Logger,
        auth_user_id_supplier: Optional[Callable[[], Optional[int]]] = None,
        *,
        max_attempts: int = 10,
        retry_sleep_sec: int = 15,
    ) -> None:
        """
        Инициализирует асинхронный транспорт.

        :param meta_url: Базовый URL META API (например, https://apimeta.devision.io).
        :param default_headers: Словарь заголовков по умолчанию для всех запросов.
        :param logger: Экземпляр логгера для записи предупреждений и ошибок.
        :param auth_user_id_supplier: Опциональный callable,
            возвращающий ID пользователя для заголовка X-META-AuthUserID.
        :param max_attempts: Максимальное количество попыток при ошибках (по умолчанию 10).
        :param retry_sleep_sec: Время ожидания между попытками в секундах (по умолчанию 15).

        :return: None.
        """
        self._meta_url = meta_url.rstrip("/")
        self._default_headers = dict(default_headers or {})
        self._log = logger
        self._auth_user_id_supplier = auth_user_id_supplier
        self._max_attempts = max_attempts
        self._retry_sleep_sec = retry_sleep_sec
        self._client = httpx.AsyncClient()

    async def aclose(self) -> None:
        """
        Закрывает HTTP-клиент и освобождает ресурсы.

        Обязательно вызывайте этот метод по завершении работы с транспортом,
        чтобы корректно закрыть все TCP-соединения.

        :return: None.
        """
        await self._client.aclose()

    async def call(
        self,
        service: str,
        method: str,
        data: Optional[dict[str, Any]],
        options: Optional[dict[str, Any]],
        multipart_form: bool = False,
        multipart_form_data: Optional[dict[str, Any]] = None,
        stream: bool = False,
        http_path: str = "/api/meta/v1/",
        http_method: str = "POST",
        get_params: Optional[dict[str, Any]] = None,
        connect_timeout_sec: int = 60,
        request_timeout_sec: int = 1800,
    ) -> httpx.Response:
        """
        Выполняет асинхронный HTTP-запрос к META API.

        :param service: Название сервиса (например, "db", "metaql").
        :param method: Метод сервиса (например, "query", "update").
        :param data: Данные запроса (тело).
        :param options: Дополнительные опции, которые будут добавлены в тело запроса.
        :param multipart_form: Флаг отправки данных в формате multipart/form-data.
        :param multipart_form_data: Файлы для multipart-запроса.
        :param stream: Флаг потоковой передачи ответа.
        :param http_path: HTTP-путь к API (по умолчанию "/api/meta/v1/").
        :param http_method: HTTP-метод (по умолчанию "POST").
        :param get_params: GET-параметры запроса.
        :param connect_timeout_sec: Таймаут подключения в секундах (по умолчанию 60).
        :param request_timeout_sec: Таймаут запроса в секундах (по умолчанию 1800).
        :return: httpx.Response с ответом сервера.
        :raises ServerError: При исчерпании попыток из-за ошибок сервера.
        :raises RequestError: При ошибках клиента (4xx).
        :raises AuthError: При ошибке авторизации (401).
        """
        payload: dict[str, Any] = dict(data or {})
        if "self" in payload:
            payload.pop("self")
        if options:
            payload.update(options)

        params = dict(get_params or {})
        headers = self._build_headers()
        url = self._compose_url(http_path, service, method)

        request_context = {
            "url": url,
            "method": http_method,
            "headers": headers,
            "params": params,
            "timeout": (connect_timeout_sec, request_timeout_sec),
            "stream": stream,
        }

        body_kwargs: dict[str, Any]
        if multipart_form:
            headers.pop("content-type", None)
            body_kwargs = {"data": payload}
            if multipart_form_data:
                body_kwargs["files"] = multipart_form_data
        else:
            body_kwargs = {"content": json.dumps(payload)}

        timeout = httpx.Timeout(request_timeout_sec, connect=connect_timeout_sec)
        last_exception: Exception = ServerError(request_context)

        for _ in range(self._max_attempts):
            try:
                request = self._client.build_request(
                    http_method,
                    url,
                    headers=dict(headers),
                    params=params,
                    **body_kwargs,
                )
                request.extensions["timeout"] = timeout.as_dict()
                response = await self._client.send(request, stream=stream)
                self._log.set_entity("request_id", response.headers.get("request_id"))
                if response.status_code == 200:
                    return response

                await response.aread()
                process_meta_api_error_code(response.status_code, request_context, response.text)
            except (httpx.TransportError, ConnectionError, TimeoutError) as exc:
                self._log.warning("META API Connection Error. Sleep...", {"e": exc})
                await asyncio.sleep(self._retry_sleep_sec)
            except ServerError as exc:
                last_exception = exc
                self._log.warning("META Server Error. Sleep...", {"e": exc})
                await asyncio.sleep(self._retry_sleep_sec)
            except Exception as exc:
                if "Служба частично или полностью недоступна" in str(exc):
                    self._log.warning("META API Service Temporarily Unavailable. Sleep...", {"e": exc})
                    await asyncio.sleep(self._retry_sleep_sec)
                else:
                    raise
            finally:
                self._log.set_entity("request_id", None)

        raise last_exception

    def _build_headers(self) -> dict[str, str]:
        """
        Формирует заголовки для запроса.

        Добавляет X-META-AuthUserID, если задан auth_user_id_supplier.

        :return: Словарь заголовков.
        """
        headers = dict(self._default_headers)
        auth_user_id = self._auth_user_id_supplier() if self._auth_user_id_supplier else None
        if auth_user_id:
            headers["X-META-AuthUserID"] = str(auth_user_id)
        return headers

    def _compose_url(self, http_path: str, service: str, method: str) -> str:
        """
        Собирает полный URL для запроса.

        :param http_path: Базовый путь API.
        :param service: Название сервиса.
        :param method: Метод сервиса.
        :return: Полный URL.
        """
        path = http_path or "/"
        if not path.endswith("/"):
            path += "/"
        if not path.startswith("/"):
            path = "/" + path
        return f"{self._meta_url}{path}{service}/{method}"



from __future__ import annotations

import json
import os
import platform
import socket
import sys
import time
from typing import Optional

import requests
from urllib3.connection import HTTPConnection

from metasdk.__state import set_current_app
from metasdk.event_bus import EventBus
from metasdk.exceptions import UnexpectedError, DbQueryError, ServerError
from metasdk.internal import read_developer_settings
from metasdk.logger import create_logger, eprint
from metasdk.logger.bulk_logger import BulkLogger
from metasdk.logger.logger import Logger
from metasdk.services import get_api_call_headers, process_meta_api_error_code
from metasdk.services.ApiProxyService import ApiProxyService
from metasdk.services.AsyncDbQueryService import AsyncDbQueryService
from metasdk.services.AuthService import AuthService
from metasdk.services.CacheService import CacheService
from metasdk.services.DbQueryService import DbQueryService
from metasdk.services.DbService import DbService
from metasdk.services.DevService import DevService
from metasdk.services.ExportService import ExportService
from metasdk.services.ExternalSystemService import ExternalSystemService
from metasdk.services.FeedService import FeedService
from metasdk.services.IssueService import IssueService
from metasdk.services.MediaService import MediaService
from metasdk.services.MetaqlService import MetaqlService
from metasdk.services.ObjectLogService import ObjectLogService
from metasdk.services.SettingsService import SettingsService
from metasdk.services.UserManagementService import UserManagementService
from metasdk.services.StarterService import StarterService
from metasdk.services.MailService import MailService
from metasdk.services.LockService import LockService
from metasdk.transport import AsyncNativeApiTransport
from metasdk.worker import Worker

DEV_STARTER_STUB_URL = "http://STUB_URL"


class MetaApp(object):
    current_app: Optional["MetaApp"] = None
    debug: bool = False
    service_id: Optional[str] = None
    build_num: Optional[str] = None
    starter_api_url: Optional[str] = None
    meta_url: Optional[str] = None
    redis_url: Optional[str] = None
    api_proxy_url: Optional[str] = None
    log: Logger = Logger()
    worker: Optional[Worker] = None
    event_bus: Optional[EventBus] = None

    # Будет поставляться в конец UserAgent
    user_agent_postfix = ""

    developer_settings = None

    # Пользователь, из под которого пройдет авторизация после того,
    # как мета авторизует разработчика, в случае, если разработчик имеет разрешения для авторизации из-под других пользователей
    auth_user_id: Optional[int] = None

    MediaService: Optional[MediaService] = None
    MetaqlService: Optional[MetaqlService] = None
    ExportService: Optional[ExportService] = None
    CacheService: Optional[CacheService] = None
    SettingsService: Optional[SettingsService] = None
    IssueService: Optional[IssueService] = None
    UserManagementService: Optional[UserManagementService] = None
    StarterService: Optional[StarterService] = None
    MailService: Optional[MailService] = None
    ApiProxyService: Optional[ApiProxyService] = None
    AuthService: Optional[AuthService] = None

    __default_headers: dict[str, str] = {}
    __db_list: dict[str, DbQueryService] = {}
    __async_db_list: dict[str, AsyncDbQueryService] = {}
    __async_transport: Optional[AsyncNativeApiTransport] = None

    def __init__(self, service_id: Optional[str] = None, debug: Optional[bool] = None,
                 starter_api_url: Optional[str] = None,
                 meta_url: Optional[str] = None,
                 api_proxy_url: Optional[str] = None,
                 include_worker: Optional[bool] = None,
                 redis_url: Optional[str] = None,
                 setup_sockets: bool = True,
                 gcloud_log_host_port: Optional[str] = None,
                 auth_service_host: Optional[str] = None,
                 ):
        if setup_sockets:
            self.setup_sockets()

        if debug is None:
            is_prod = os.environ.get('PRODUCTION', False)
            debug = os.environ.get('DEBUG', not is_prod)
            if include_worker is None:
                include_worker = True
            if debug == 'false':
                debug = False
        self.debug = debug
        defaultEnvParamList = []
        if os.environ.get("REDIS_URL") is None and redis_url is None:
            defaultEnvParamList.append("REDIS_URL")
        if os.environ.get("META_URL") is None and meta_url is None:
            defaultEnvParamList.append("META_URL")
        if os.environ.get("API_PROXY_URL") is None and api_proxy_url is None:
            defaultEnvParamList.append("API_PROXY_URL")
        if os.environ.get("STARTER_URL") is None and starter_api_url is None:
            defaultEnvParamList.append("STARTER_URL")
        if os.environ.get("GCLOUD_LOG_HOST_PORT") is None and gcloud_log_host_port is None:
            defaultEnvParamList.append("GCLOUD_LOG_HOST_PORT")
        if len(defaultEnvParamList) > 0:
            self.log.warning("Пропущены обязательные переменные окружения: " + ", ".join(defaultEnvParamList))

        self.redis_url = os.environ.get("REDIS_URL", redis_url or "s1.meta.vmc.loc:6379:1")
        self.meta_url = os.environ.get("META_URL", meta_url or "https://apimeta.devision.io")
        self.api_proxy_url = os.environ.get("API_PROXY_URL", api_proxy_url or "https://apiproxy.apis.garpun.com")
        self.auth_service_host = os.environ.get("AUTH_SERVICE_HOST", auth_service_host or "https://account.garpun.com")

        if debug and not starter_api_url:
            starter_api_url = DEV_STARTER_STUB_URL
        self.starter_api_url = os.environ.get("STARTER_URL", starter_api_url or "http://starter_url_not_set")

        if service_id:
            self.log.warning("Параметр service_id скоро будет удален из MetaApp")

        gcloud_log_host_port = os.environ.get("GCLOUD_LOG_HOST_PORT", gcloud_log_host_port or "n3.adp.vmc.loc:31891")
        service_ns = os.environ.get('SERVICE_NAMESPACE', "appscript")  # для ns в логах и для префикса Dispatcher
        service_id = os.environ.get('SERVICE_ID', "local_debug_service")
        self.build_num = os.environ.get('BUILD_NUM', '0')
        self.service_ns = service_ns
        self.service_id = service_id
        self.dispatcher_name = self.service_ns + "." + self.service_id

        create_logger(service_id=service_id, service_ns=service_ns, build_num=self.build_num,
                      gcloud_log_host_port=gcloud_log_host_port, debug=self.debug)

        self.__read_developer_settings()

        self.__default_headers = get_api_call_headers(self)

        # Оставляем синхронный sync-кеш (__db_list) как было (классовым),
        # чтобы не ломать старый функционал. Асинхронные кеши/транспорт делаем инстансовыми,
        # чтобы не делить закрытый клиент между MetaApp.
        self.__async_db_list = {}

        self.__async_transport = None
        self.AuthService = AuthService(self)
        self.MediaService = MediaService(self)
        self.MetaqlService = MetaqlService(self)
        self.SettingsService = SettingsService(self)
        self.ExportService = ExportService(self)
        self.CacheService = CacheService(self)
        self.IssueService = IssueService(self)
        self.StarterService = StarterService(self, self.db("meta"), self.starter_api_url)
        self.MailService = MailService(self)
        self.DbService = DbService(self)
        self.UserManagementService = UserManagementService(self)
        self.ApiProxyService = ApiProxyService(self)
        self.ExternalSystemService = ExternalSystemService(self)
        self.FeedService = FeedService(self)
        self.LockService = LockService(self)
        self.ObjectLogService = ObjectLogService(self)
        self.DevService = DevService(self)

        if include_worker:
            self.event_bus = EventBus(self)

            stdin = "[]" if debug else ''.join(sys.stdin.readlines())
            self.worker = Worker(self, stdin)

        set_current_app(self)
        self.log.info("MetaApp initialized", {"version": self.get_lib_version()})

    def bulk_log(self, log_message=u"Еще одна пачка обработана", total=None, part_log_time_minutes=5):
        """
        Возвращает инстант логгера для обработки списокв данных
        :param log_message: То, что будет написано, когда время придет
        :param total: Общее кол-во объектов, если вы знаете его
        :param part_log_time_minutes: Раз в какое кол-во минут пытаться писать лог
        :return: BulkLogger
        """
        return BulkLogger(log=self.log, log_message=log_message, total=total,
                          part_log_time_minutes=part_log_time_minutes)

    def db(self, db_alias, shard_key=None):
        """
        Получить экземпляр работы с БД
        :type db_alias: basestring Альяс БД из меты
        :type shard_key: Любой тип. Некоторый идентификатор, который поможет мете найти нужную шарду. Тип зависи от принимающей стороны
        :rtype: DbQueryService
        """
        if shard_key is None:
            shard_key = ''

        db_key = db_alias + '__' + str(shard_key)
        if db_key not in self.__db_list:
            self.__db_list[db_key] = DbQueryService(self, {"db_alias": db_alias, "dbAlias": db_alias,
                                                           "shard_find_key": shard_key, "shardKey": shard_key})
        return self.__db_list[db_key]

    def db_async(self, db_alias: str, shard_key: Optional[str] = None) -> AsyncDbQueryService:
        """
        Получить экземпляр асинхронного клиента, использующего общий кешированный транспорт.

        Подходит для долгоживущих приложений/воркеров: транспорт создаётся один раз и
        не закрывается автоматически. После завершения работы приложения следует вызвать
        `await app.close_async_transport()`. Конечно, после закрытия контейнера ОС и так зачистит сокеты,
        однако во избежание все же рекомендуется вызывать метод закрытия транспорта.
        :param db_alias: Алиас БД.
        :param shard_key: Название ключа шарда (опционально).
        :return: AsyncDbQueryService.
        """
        if shard_key is None:
            shard_key = ""

        db_key = f"{str(db_alias)}___{str(shard_key)}"
        if db_key not in self.__async_db_list:
            self.__async_db_list[db_key] = AsyncDbQueryService(
                self,
                {
                    "db_alias": db_alias,
                    "dbAlias": db_alias,
                    "shard_find_key": shard_key,
                    "shardKey": shard_key
                },
            )
        return self.__async_db_list[db_key]

    def db_async_session(self, db_alias: str, shard_key: Optional[str] = None) -> AsyncDbQueryService:
        """
        Создаёт асинхронный клиент с выделенным транспортом под контекст.

        Использование: `async with app.db_async_session(...) as db:`. Транспорт будет
        закрыт автоматически при выходе из контекста и не влияет на общий кешированный
        транспорт.
        """
        if shard_key is None:
            shard_key = ""

        transport = self._create_async_transport()
        return AsyncDbQueryService(
            self,
            {
                "db_alias": db_alias,
                "dbAlias": db_alias,
                "shard_find_key": shard_key,
                "shardKey": shard_key,
            },
            transport=transport,
            owns_transport=True,
        )

    @property
    def user_agent(self):
        return self.service_id + " | b" + self.build_num + (
            ' | ' + self.user_agent_postfix if self.user_agent_postfix else "")

    def __read_developer_settings(self):
        """
        Читает конфигурации разработчика с локальной машины или из переменных окружения
        При этом переменная окружения приоритетнее
        :return:
        """
        self.developer_settings = read_developer_settings()
        if not self.developer_settings:
            self.log.warning(
                "НЕ УСТАНОВЛЕНЫ настройки разработчика, это может приводить к проблемам в дальнейшей работе!")

    def api_call(self, service, method, data, options):
        """
        :type app: metasdk.MetaApp
        """
        if 'self' in data:
            # может не быть, если вызывается напрямую из кода,
            # а не из прослоек типа DbQueryService
            data.pop("self")

        if options:
            data.update(options)

        _headers = dict(self.__default_headers)

        if self.auth_user_id:
            _headers['X-META-AuthUserID'] = str(self.auth_user_id)

        request = {
            "url": self.meta_url + "/api/v1/adptools/" + service + "/" + method,
            "data": json.dumps(data),
            "headers": _headers,
            "timeout": (60, 1800)
        }

        last_e = ServerError(request)
        for _try_idx in range(20):
            try:
                resp = requests.post(**request)
                if resp.status_code == 200:
                    decoded_resp = json.loads(resp.text)
                    if 'data' in decoded_resp:
                        return decoded_resp['data'][method]
                    if 'error' in decoded_resp:
                        if 'details' in decoded_resp['error']:
                            eprint(decoded_resp['error']['details'])
                        raise DbQueryError(decoded_resp['error'])
                    raise UnexpectedError()
                else:
                    process_meta_api_error_code(resp.status_code, request, resp.text)
            except (requests.exceptions.ConnectionError, ConnectionError, TimeoutError) as e:
                self.log.warning('META API Connection Error. Sleep...', {"e": e})
                time.sleep(15)

            except ServerError as e:
                last_e = e
                self.log.warning('META Server Error. Sleep...', {"e": e})
                time.sleep(15)

            except Exception as e:
                if 'Служба частично или полностью недоступна' in str(e):
                    self.log.warning('META API Connection Error. Sleep...', {"e": e})
                    time.sleep(15)
                else:
                    raise e

        raise last_e

    def native_api_call(self, service, method, data, options, multipart_form=False, multipart_form_data=None,
                        stream=False, http_path="/api/meta/v1/", http_method='POST',
                        get_params=None, connect_timeout_sec=60, request_timeout_sec=1800):
        """
        :type app: metasdk.MetaApp
        :rtype: requests.Response
        """
        if get_params is None:
            get_params = {}
        if 'self' in data:
            # может не быть, если вызывается напрямую из кода,
            # а не из прослоек типа DbQueryService
            data.pop("self")

        if options:
            data.update(options)

        _headers = dict(self.__default_headers)

        if self.auth_user_id:
            _headers['X-META-AuthUserID'] = str(self.auth_user_id)

        request = {
            "url": self.meta_url + http_path + service + "/" + method,
            "timeout": (connect_timeout_sec, request_timeout_sec),
            "stream": stream,
            "params": get_params,
        }

        if multipart_form:
            if multipart_form_data:
                request['files'] = multipart_form_data
            request['data'] = data
            _headers.pop('content-type', None)
        else:
            request['data'] = json.dumps(data)
        request['headers'] = _headers

        last_e = ServerError(request)
        for _try_idx in range(10):
            try:
                resp = requests.request(http_method, **request)
                # добавляем глобальную трассировку в логи
                self.log.set_entity("request_id", resp.headers.get('request_id'))
                if resp.status_code == 200:
                    return resp
                else:
                    process_meta_api_error_code(resp.status_code, request, resp.text)
            except (requests.exceptions.ConnectionError, ConnectionError, TimeoutError) as e:
                self.log.warning('META API Connection Error. Sleep...', {"e": e})
                time.sleep(15)

            except ServerError as e:
                last_e = e
                self.log.warning('META Server Error. Sleep...', {"e": e})
                time.sleep(15)

            except Exception as e:
                if 'Служба частично или полностью недоступна' in str(e):
                    self.log.warning('META API Service Temporarily Unavailable. Sleep...', {"e": e})
                    time.sleep(15)
                else:
                    raise e
            finally:
                self.log.set_entity("request_id", None)

        raise last_e

    def __ensure_async_transport(self) -> AsyncNativeApiTransport:
        """
        Убеждается, что асинхронный транспорт инициализирован.

        :return: AsyncNativeApiTransport.
        """
        if not self.__async_transport:
            if not self.meta_url:
                raise ValueError("MetaApp.meta_url must be configured before using async transport")
            self.__async_transport = AsyncNativeApiTransport(
                meta_url=self.meta_url,
                default_headers=self.__default_headers,
                logger=self.log,
                auth_user_id_supplier=lambda: self.auth_user_id,
            )
        return self.__async_transport

    def ensure_async_transport(self) -> AsyncNativeApiTransport:
        """
        Публичный доступ к кешированному асинхронному транспорту.
        """
        return self.__ensure_async_transport()

    def _create_async_transport(self) -> AsyncNativeApiTransport:
        """
        Создаёт новый экземпляр асинхронного транспорта без кеширования.
        """
        if not self.meta_url:
            raise ValueError("MetaApp.meta_url must be configured before using async transport")
        return AsyncNativeApiTransport(
            meta_url=self.meta_url,
            default_headers=self.__default_headers,
            logger=self.log,
            auth_user_id_supplier=lambda: self.auth_user_id,
        )

    async def close_async_transport(self) -> None:
        """
        Закрывает асинхронный транспорт.

        :return: None.
        """
        if self.__async_transport:
            await self.__async_transport.aclose()
            self.__async_transport = None

    def get_lib_version(self) -> str:
        """
        Возвращает версию библиотеки MetaSDK.

        :return: Версия библиотеки.
        """
        from metasdk import info
        return info.__version__

    @staticmethod
    def setup_sockets():
        """
        Настройка для решения проблемы с разрывом соединения после, примерно, трёх минут при запросе к сервисам,
        расположенным ва яндекс-облаке
        (больше информации в задаче META-2932)
        """

        # Не все из этих параметров есть в каждой системе (в MacOS, например, нет socket.TCP_KEEPIDLE).
        # Если сделать обычные списки, то при интерпретации отсутствующего параметра возникнет исключение.
        # Чтобы этого избежать всё завёрнуто в лямбды
        additional_socket_options_for_systems = {
            "Windows": lambda: [(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
                                (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 10),
                                (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 60),
                                (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)],
            "Linux": lambda: [(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
                              (socket.IPPROTO_TCP, socket.TCP_KEEPIDLE, 1),
                              (socket.IPPROTO_TCP, socket.TCP_KEEPINTVL, 3),
                              (socket.IPPROTO_TCP, socket.TCP_KEEPCNT, 5)],
            "Darwin": lambda: [(socket.SOL_SOCKET, socket.SO_KEEPALIVE, 1),
                               (socket.IPPROTO_TCP, 0x10, 3)]
        }

        HTTPConnection.default_socket_options += additional_socket_options_for_systems[platform.system()]()



import json

from abc import ABC, abstractmethod
from datetime import datetime
from typing import Any, Mapping, Optional, TYPE_CHECKING
from uuid import UUID


if TYPE_CHECKING:
    from metasdk.MetaApp import MetaApp  # type: ignore


class Message(ABC):
    """Абстрактный класс сообщения"""
    messenger_id = "abstract"

    @abstractmethod
    def generate_unique_id(self) -> str:
        """
        Генерирует уникальный идентификатор сообщения по его атрибутам.

        Абстрактный метод. В наследниках необходимо переопределить.
        :return: Уникальный идентификатор сообщения.
        """
        pass

    def __init__(
        self,
        user_id: str,
        body: str,
        template: str = "",
        unique_id: Optional[str] = None,
        author_user_id: Optional[int] = None,
        sender_id: Optional[str] = None,
        message_extra: Optional[Mapping[str, Any]] = None,
        **kwargs: Any,
    ) -> None:
        """
        Инициализирует экземпляр класса.

        :param user_id: Идентификатор пользователя-адресата.
        :param body: Строка с телом отправляемого сообщения.
        :param template: Идентификатор шаблона сообщения.
        :param unique_id: Уникальный идентификатор сообщения.
        :param author_user_id: Идентификатор автора сообщения.
        :param sender_id: Идентификатор отправителя.
        :param message_extra: Дополнительные параметры сообщения.
        :param kwargs: Опциональные параметры, актуальные для отдельных каналов отправки сообщений.
        :return: None.
        """
        self._messenger = self.messenger_id
        self._user_id = user_id
        self._body = body
        self._template = template
        self._custom_unique_id = unique_id
        self._author_user_id = author_user_id
        self._sender_id = sender_id
        self._message_extra = message_extra
        self._kwargs = kwargs

    def to_dict(self) -> dict[str, str | int | None]:
        """
        Возвращает словарь с данными сообщения.

        :return: Словарь с данными сообщения. Удобно передавать методам DbQueryService.
        """
        sender_info = {}
        if self._sender_id:
            sender_info["id"] = self._sender_id

        if self._message_extra:
            message_extra = self._message_extra
        else:
            message_extra = {}

        return {
            "messenger": self._messenger,
            "user_id": self._user_id,
            "body": self._body,
            "template": self._template,
            "unique_id": self._custom_unique_id or self.generate_unique_id(),
            "author_user_id": self._author_user_id,
            "sender_info": json.dumps(sender_info),
            "message_extra": json.dumps(message_extra),
        }


class WebhookMessage(Message):
    """Класс сообщения, отправляемого вебхуком"""
    messenger_id = "webhook"

    def generate_unique_id(self) -> str:
        """
        Генерирует уникальный идентификатор сообщения по его атрибутам.

        В kwargs экземпляра можно передать параметр "uniqualizer",
        и он будет добавлен в начало unique_id. Так, скажем, можно проверять,
        был ли уже отправлен вебхук по конкретной записи из meta.object_log. 
        :return: Уникальный идентификатор сообщения.
        """
        unique_id_head = "webhook"
        if uniqualizer := self._kwargs.get("uniqualizer"):
            unique_id_head += f"_{uniqualizer}"       
        timestamp = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        return f"{unique_id_head}_{self._user_id}_{timestamp}"


class TelegramMessage(Message):
    """Класс сообщения, отправляемого через Telegram"""
    messenger_id = "telegram"

    def __init__(
        self, 
        user_id: str, 
        body: str, 
        sender_id: str = "80a5b860-683e-4afd-9733-572826bbac1e",  # https://t.me/Garpun_bot
        template: str = "", 
        unique_id: Optional[str] = None, 
        author_user_id: Optional[int] = None, 
        **kwargs: Any,
    ) -> None:
        """
        Инициализирует экземпляр класса.

        :param user_id: Идентификатор пользователя-адресата.
        :param body: Строка с телом отправляемого сообщения.
        :param sender_id: Идентификатор отправителя (бота в таблице bot.telegram_bot).
        :param template: Идентификатор шаблона сообщения.
        :param unique_id: Уникальный идентификатор сообщения.
        :param author_user_id: Идентификатор автора сообщения.
        :param message_extra: Дополнительные параметры сообщения. Примеры таких параметров для Telegram:
            - "inline_buttons" (list) — список кнопок для встраивания в сообщение;
            - "keep_formatting" (bool) — если True, форматирование сообщения будет сохранено в неизменном виде;
            - "ignore_queue" (bool) — если True, сообщение не будет добавлено в очередь на отправку.
            - "parse_mode" (str) — режим парсинга сообщения на стороне Telegram (по умолчанию "HTML").
            - "image_url" (str) — URL изображения для встраивания в сообщение.
            - "sticker_id" (str) — ID стикера для встраивания в сообщение. Если передан, будет отправлен только стикер,
                так как, в отличие от изображений, к стикеру нельзя прикрепить текст.
        :param kwargs: Опциональные параметры, актуальные для отдельных каналов отправки сообщений.
        :return: None.
        """
        try:
            UUID(sender_id)
        except ValueError:
            raise ValueError("Invalid sender_id")
        super().__init__(user_id, body, template, unique_id, author_user_id, sender_id, **kwargs)

    def generate_unique_id(self) -> str:
        """
        Генерирует уникальный идентификатор сообщения по его атрибутам.

        :return: Уникальный идентификатор сообщения.
        """
        timestamp = datetime.now().strftime("%Y-%m-%d_%H:%M:%S")
        return f"custom_message_{self._user_id}_{timestamp}_{self._sender_id}"


class MailService:
    def __init__(self, app: "MetaApp") -> None:
        """
        Инициализирует экземпляр класса MailService.

        :param app: Экземпляр приложения MetaApp.
        :return: None.
        """
        self.__app = app
        self.__options: dict[str, Any] = {}
        self.__data_get_cache: dict[str, Any] = {}
        self.__metadb = app.db("meta")
        self.log = app.log

    def submit_mail(
        self,
        send_from: str,
        send_to: str,
        subject: str,
        body: str,
        unique_id: Optional[str] = None,
    ) -> None:
        """
        Добавляет письмо в очередь на отправку (фактически — запись в таблицу meta.mail).

        :param send_from: Отправитель.
        :param send_to: Получатель.
        :param subject: Тема письма.
        :param body: Тело письма. Можно с HTML.
        :param unique_id: Уникальный идентификатор письма.
            Лучше всего подойдет md5 + человекочитаемый префикс.
            Письмо с существующим unique_id не будут добавлено.
        :return: None.
        """
        self.__metadb.update(
            """
                INSERT INTO meta.mail(
                    "template",
                    "from",
                    "to",
                    "subject",
                    "body",
                    "attachments",
                    "unique_id"
                )
                VALUES ('meta', :send_from, :send_to, :subject, :body, null, :unique_id)
                ON CONFLICT (unique_id) DO NOTHING
            """,
            {
                "send_from": send_from,
                "send_to": send_to,
                "subject": subject,
                "body": body,
                "unique_id": unique_id,
            },
        )

    def submit_message(self, message: Message) -> str | None:
        """
        Добавляет сообщение в очередь на отправку (фактически — запись в таблицу meta.messenger).

        :param message: Экземпляр класса сообщения (наследника Message).
        :return: id записи в таблице meta.messenger или None, если не удалось добавить запись.
        """
        insert_query_result = self.__metadb.one(
            """
                INSERT INTO meta.messenger(
                    messenger,
                    user_id,
                    body,
                    unique_id,
                    author_user_id, 
                    sender_info,
                    message_extra
                )
                VALUES (
                    :messenger::text,
                    :user_id::text,
                    :body::text,
                    :unique_id::text,
                    COALESCE(:author_user_id, valera_user_id())::bigint,
                    :sender_info::jsonb,
                    :message_extra::jsonb
                )
                ON CONFLICT (unique_id) DO NOTHING
                RETURNING id
            """,
            message.to_dict(),
        )
        return insert_query_result.get("id") if insert_query_result else None


class TelegramUtils:
    """Класс вспомогательных методов для работы с Telegram"""
    def __init__(self, app: "MetaApp") -> None:
        """
        Инициализирует экземпляр класса.

        :param app: Экземпляр приложения MetaApp.
        :return: None.
        """
        self.__app = app
        self.__metadb = self.__app.db("meta")
        self.__log = self.__app.log

    def get_bot_token(self, bot_id: str) -> str | None:
        """
        Получает токен бота по его идентификатору.

        :param bot_id: Идентификатор бота.
        :return: Токен бота либо None, если бот не найден.
        """
        token_query_result = self.__metadb.one(
            """
            SELECT token
            FROM bot.telegram_bot
            WHERE id = :bot_id::uuid
            LIMIT 1
            """,
            {"bot_id": bot_id},
        )
        return token_query_result.get("token") if token_query_result else None

    def submit_message_deprecated(self, user_id: int, msg: str) -> bool:
        """
        Отправляет уведомление пользователю в Telegram от @RealwebIntranetBot.

        С отключением сервиса бота функция потеряет актуальность и будет удалена.
        Не рекомендуется использовать в новых службах и сервисах.
        :param user_id: ID пользователя в META.
        :param msg: Сообщение для отправки.
        :return: True, если сообщение отправлено успешно, иначе False.
        """
        notification_result = self.__metadb.one(
            """
            SELECT *
            FROM api.telegram_bot_rw_request(
                'message/send',
                json_build_object(
                    'userId', (SELECT intranet_id FROM public.users WHERE id=:user_id::bigint LIMIT 1),
                    'messages', json_build_array(json_build_object('text', :msg))
                )
            )
            """,
            {"user_id": user_id, "msg": msg},
        )

        if "DATA_ERROR" in str(notification_result):
            self.__log.warning(
                "При отправке сообщения произошла ошибка",
                {"user_id": user_id, "result": notification_result},
            )
            return False
        else:
            self.__log.info(
                "Сообщение от RealwebIntranetBot отправлено",
                {"user_id": user_id, "result": notification_result},
            )
            return True

    def fetch_user_telegram_id(
        self,
        user_id: Optional[int] = None,
        user_email: Optional[str] = None,
        strict: bool = False,
    ) -> int | None:
        """
        Получает Telegram ID пользователя по его ID в META или e-mail.

        :param user_id: ID пользователя в META (В первую очередь поиск идет по нему).
        :param user_email: E-mail пользователя (Вторая очередь поиска, только по юзерам с company_id = 1).
        :param strict: Флаг, определяющий режим поиска (По умолчанию False).
            Если True и переданы оба параметра, проверка будет производиться на соответствие обоим.
            Если False и переданы оба параметра, будет возвращен первый найденный Telegram ID.
        :return: Telegram ID пользователя либо None, если пользователь не найден.
        """
        if user_id is None and user_email is None:
            raise ValueError("Необходимо передать хотя бы один из параметров: user_id или user_email")
        target_telegram_id = None

        if user_id is not None:
            user_by_id_query = self.__metadb.one(
                """
                SELECT
                    pu.id AS meta_id,
                    pu.telegram_id AS telegram_id,
                    pu.email AS meta_email,
                    CONCAT(pu.info->'intranet'->>'domain_id', '@realweb.ru') AS domain_email
                FROM public.users pu
                WHERE id = :user_id::bigint
                """,
                {"user_id": user_id},
            )
            target_telegram_id = user_by_id_query.get("telegram_id") if user_by_id_query else None
            if target_telegram_id and not strict:
                return target_telegram_id

        # Сюда мы попадаем, если user_id не передан или по нему не найден Telegram ID либо при строгом поиске
        if user_email is not None:
            users_by_email_query = self.__metadb.all(
                """
                SELECT
                    pu.id AS meta_id,
                    pu.telegram_id AS telegram_id,
                    pu.email AS meta_email,
                    CONCAT(pu.info->'intranet'->>'domain_id', '@realweb.ru') AS domain_email
                FROM public.users pu
                WHERE
                    (
                        CONCAT(pu.info->'intranet'->>'domain_id', '@realweb.ru') = :user_email::text
                        OR pu.email = :user_email::text
                    )
                    AND pu.company_id = 1
                """,
                {"user_email": user_email},
            )
            if strict and target_telegram_id:
                if target_telegram_id in [user["telegram_id"] for user in users_by_email_query]:
                    return target_telegram_id
                else:
                    return None
            elif not strict:
                results_with_telegram_id = [result for result in users_by_email_query if result.get("telegram_id")]
                if results_with_telegram_id:
                    return results_with_telegram_id[0].get("telegram_id")
                else:
                    return None
        # Если вообще ничего не найдено, возвращаем None
        return None

    def is_bot_subscriber(self, bot_id: str, telegram_id: int) -> bool:
        """
        Проверяет, подписан ли пользователь на бота.

        Актуально только для ботов бот-платформы, для которых собирается информация о подписчиках.
        :param bot_id: Идентификатор бота.
        :param telegram_id: Telegram ID пользователя.
        :return: True, если пользователь подписан на бота, иначе False.
        """
        is_subscriber_query_result = self.__metadb.one(
            """
            SELECT
                EXISTS (
                    SELECT 1
                    FROM bot.subscriber
                    WHERE
                        bot_id = :bot_id::uuid
                        AND status = 'active'
                        AND remote_id=:telegram_id::bigint
                ) AS is_bot_subscriber
            """,
            {"telegram_id": telegram_id, "bot_id": bot_id},
        )
        return is_subscriber_query_result.get("is_bot_subscriber", False) if is_subscriber_query_result else False
